#!/usr/local/bin/python

# igen.py -- Automatic interface generator for Terry.
#
# This file is part of Terry, a Python binding for RenderMan.
# 
# Copyright (C) 1996, 1997 Ture Plsson.
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
# 
# The RenderMan (R) Interface Procedures and RIB Protocol are:
#     Copyright 1988, 1989, Pixar.  All rights reserved.
# RenderMan (R) is a registered trademark of Pixar.

import sys, string

template = """
PyObject *
terry_py_%s(PyObject *self, PyObject *pyargs, PyObject *kwdargs)
{
    %s
    %s
    
    RtToken *tokens;
    RtPointer *values;
    int nparams;
    PyObject *dict;

    dict = 0;
    if (!terry_parse_args(%d, pyargs, %s, argsyntax, %s))
        return 0;

    if (!terry_parse_dict_and_kwd(dict, kwdargs, &nparams, &tokens, &values))
        return 0;
    
    %s %s(%s %s);

    terry_free_parameter_list(nparams, tokens, values);
    %s
}
"""

bindings_prologue = ''

table_prologue = '''static PyMethodDef terry_methods[] = {
    { "setsymdict",  	terry_py_setsymdict,   METH_VARARGS },
    { "declare_c",      terry_py_declare_c,    METH_VARARGS },
    { "polygon",        terry_py_polygon,      METH_VARARGS | METH_KEYWORDS },
    { "nupatch",        terry_py_nupatch,      METH_VARARGS | METH_KEYWORDS },
'''

table_epilogue = '''    { NULL, NULL }
};
'''

def parsetype(name):
    # NOTE: This had better be in the same order as the enum in terry.h
    typenames = [ 'RtBoolean', 'RtInt', 'RtFloat', 'RtString', 'RtToken',
		  'RtPoint', 'RtColor', 'RtMatrix', 'RtBasis', 'RtBound',
		  'RtLightHandle', 'RtObjectHandle', 'RtFilterFunc', 'RtFunc',
		  'RtPointer' ]

    if len(name) > 2 and name[-2:] == '[]':
	return (0, typenames.index(name[0:-2]), -1)
    else:
	return (0, typenames.index(name), 0)

def genfunction(outf, fdesc):
    rettype = fdesc[0]
    name = fdesc[1]

    if len(fdesc) > 2 and fdesc[-1] == '...':
	variadic = 1
	args = fdesc[2: -1]
	dictp = '&dict'
    else:
	variadic = 0
	args = fdesc[2:]
	dictp = '(PyObject**)0'

    # Remove those args that exist only in rhe RIB binding, and counters.
    i = 0
    pos = 0
    counters = {}
    while i < len(args):
	if args[i][0] == '#':
	    args[i : i + 2] = []
	elif args[i][0] == '%':
	    counters[args[i][1:]] = pos
	    args[i : i + 2] 
	    pos = pos + 1
	else:
	    pos = pos + 1
	    
	i = i + 1

    if variadic:
	name = 'Ri' + name + 'V'
	if len(args) > 0:
	    calltail = ', nparams, tokens, values'
	else:
	    calltail = 'nparams, tokens, values'
    else:
	name = 'Ri' + name
	calltail = ''

    # Split what's left of the list into names and types
    i = 0
    names = []
    types = []
    while i < len(args) - 1:
	types.append(args[i])
	names.append(args[i + 1])
	i = i + 2

    # Parse the types for the remaining arguments, and turn them into
    # structure initializers.
    argtypes = []
    for x in types:
	try:
	    argtypes.append(parsetype(x))
	except ValueError:
	    sys.stderr.write('bad type "%s"\n' % x)
	    sys.exit(1)

    arginits = ''
    for x in argtypes:
	arginits = arginits + '{ %d, %d, %d },\n' % x

    # Build declarations for temporary storage of the fixed parameters
    argdecl = ''
    if len(names) > 0:
	for i in range(0, len(names)):
	    argdecl = argdecl + types[i] + ' ' + names[i] + ';\n'
	    i = i + 1

    # Declare a result receiver for requests that don't return void
    if rettype == 'RtLightHandle' or rettype == 'RtObjectHandle':
	argdecl = argdecl + 'void *handle;\n'
	resultrcv = 'handle = '
	doreturn = 'return PyInt_FromLong((long)handle);\n'
    else:
	resultrcv = ''
	doreturn = '    Py_INCREF(Py_None);\n    return Py_None;\n'

    # Build a list of address-of operators for the fixed parameter
    # storage
    def addampersand(x): return '&' + x
    if len(names) > 0:
	addroff = string.joinfields(map(addampersand, names), ', ')
    else:
	addroff = '0'

    # Actual parameters to Ri call
    if len(names) > 0:
	actualparams = string.joinfields(names, ', ')
    else:
	actualparams = ''

    if arginits == '':
	arginits = 'static struct type argsyntax[1];'
    else:
	arginits = 'static struct type argsyntax[] = { %s };' \
		   % (arginits,)

    # Finally, output that function
    outf.write(template % (fdesc[1], argdecl, arginits, len(names), 
			   dictp, addroff, resultrcv, name, actualparams,
			   calltail, doreturn))


bind_out = open('GENbindings.c', 'w')
tabl_out = open('GENtable.c', 'w')

bind_out.write(bindings_prologue)

interface = sys.stdin.readlines()

for l in interface:
    desc = string.split(l)
    if l[0] != ';' and desc[1] != 'version':
	genfunction(bind_out, desc)

tabl_out.write(table_prologue)
for l in interface:
    desc = string.split(l)
    if l[0] != ';' and desc[1] != 'version':
	tabl_out.write('    { "%s", terry_py_%s, METH_VARARGS | METH_KEYWORDS},\n'
		       % (string.lower(desc[1]), desc[1]))

tabl_out.write(table_epilogue)

tabl_out.close()
bind_out.close()

