/*
 * pyinterface.c -- Terry glue functions too weird to generate automatically.
 *
 * This file is part of Terry, a Python binding for RenderMan.
 * 
 * Copyright (C) 1996, 1997 Ture Plsson.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * The RenderMan (R) Interface Procedures and RIB Protocol are:
 *     Copyright 1988, 1989, Pixar.  All rights reserved.
 * RenderMan (R) is a registered trademark of Pixar.
 */

#include <Python.h>

#include <ri.h>

#include "terry.h"

PyObject *terry_error;

PyObject *
terry_py_declare_c(PyObject *self, PyObject *args)
{
    char *name, *decl;
    RtToken token;
    if (!PyArg_ParseTuple(args, "ss", &name, &decl))
	return 0;
    token = RiDeclare(name, decl);
    return PyInt_FromLong((long)token); /* FIXME */
}

PyObject *
terry_py_polygon(PyObject *self, PyObject *args, PyObject *kwdargs)
{
    PyObject *dict, *key, *value;
    RtToken *tokens;
    RtPointer *values;
    int npars;
    int npoints;
    int n1;

    dict = 0;
    PyArg_ParseTuple(args, "|O", &dict);

    /* Figure out the number of points, by finding a "P" parameter
       and look at its size */

    key = PyString_FromString("P");
    value = dict ? PyDict_GetItem(dict, key) : 0;
    if (!value) value = kwdargs ? PyDict_GetItem(kwdargs, key) : 0;
    Py_DECREF(key);
    if (!value || !PyList_Check(value))
    {
	PyErr_SetString(terry_error, "Bad \"P\" argument type");
	return 0;
    }

    npoints = PyList_Size(value);
    terry_parse_dict_and_kwd(dict, kwdargs, &npars, &tokens, &values);

    RiPolygonV(npoints, npars, tokens, values);

    terry_free_parameter_list(npars, tokens, values);

    Py_INCREF(Py_None);
    return Py_None;
}

PyObject *
terry_py_nupatch(PyObject *self, PyObject *args, PyObject *kwd)
{
    /*
     * terry.nupatch(uorder, uknots, umin, umax, vorder, vknots, vmin, vmax, 
     *               paramdict)
     *
     * The number of control points in each direction is computed
     * from the order and the number of knot values.
     */

    static struct type float_t = { 0, T_FLOAT, 0 };
    PyObject *dict, *uknotlist, *vknotlist;
    RtToken *tokens;
    RtPointer *values;
    int npars;
    int nu, uorder, nv, vorder, nuknots, nvknots;
    double umin, umax, vmin, vmax;
    RtFloat *uknots, *vknots;

    dict = 0;
    if (!PyArg_ParseTuple(args, "iOddiOdd|O",
			  &uorder, &uknotlist, &umin, &umax,
			  &vorder, &vknotlist, &vmin, &vmax,
			  &dict))
	return 0;

    if (!terry_parse_dict_and_kwd(dict, kwd, &npars, &tokens, &values))
	return 0;

    nuknots = PyList_Size(uknotlist);
    nvknots = PyList_Size(vknotlist);
    uknots = (RtFloat*)xmalloc(sizeof(RtFloat) * nuknots);
    vknots = (RtFloat*)xmalloc(sizeof(RtFloat) * nvknots);

    if (!terry_convert_list_of_values(uknotlist, &float_t, uknots,
				      terry_sizeof(&float_t))
	|| !terry_convert_list_of_values(vknotlist, &float_t, vknots,
				 terry_sizeof(&float_t)))
    {
	xfree(uknots);
	xfree(vknots);
	terry_free_parameter_list(npars, tokens, values);
	return 0;
    }

    nu = nuknots - uorder;
    nv = nvknots - vorder;

    RiNuPatchV(nu, uorder, uknots, umin, umax, nv, vorder, vknots, vmin, vmax,
	       npars, tokens, values);

    xfree(uknots);
    xfree(vknots);
    terry_free_parameter_list(npars, tokens, values);

    Py_INCREF(Py_None);
    return Py_None;
}

#include "GENbindings.c"
#include "GENtable.c"

void
init_terry(void)
{
    PyObject *m, *d;

    m = Py_InitModule("_terry", terry_methods);
    d = PyModule_GetDict(m);
    terry_error = PyString_FromString("terry.error");
    PyDict_SetItemString(d, "error", terry_error);
}
