/*
 * terry.c -- Miscellaneous support functions for Terry.
 *
 * This file is part of Terry, a Python binding for RenderMan.
 * 
 * Copyright (C) 1996, 1997 Ture Plsson.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * The RenderMan (R) Interface Procedures and RIB Protocol are:
 *     Copyright 1988, 1989, Pixar.  All rights reserved.
 * RenderMan (R) is a registered trademark of Pixar.
 */

#include <memory.h>
#include <stdio.h>
#include <stdarg.h>

#include <Python.h>
#include <ri.h>
#include "terry.h"

/* This variable holds the symbol dictionary */
static PyObject *symdict = 0;

/* The number of colour samples */
int terry_color_samples = 3;

PyObject *
terry_py_setsymdict(PyObject *self, PyObject *args)
{
    PyObject *dict;
    if (!PyArg_ParseTuple(args, "O!", &PyDict_Type, &dict))
	return 0;

    if (dict != symdict)
    {
	if (symdict) Py_DECREF(symdict);
	symdict = dict;
	Py_INCREF(symdict);
    }

    Py_INCREF(Py_None);
    return Py_None;
}

static int
lookup_type(PyObject *key, struct type *type)
{
    PyObject *value;

    if ((value = PyDict_GetItem(symdict, key)) != 0)
    {
	type->varying   = PyInt_AsLong(PyTuple_GetItem(value, 0));
	type->base      = PyInt_AsLong(PyTuple_GetItem(value, 1));
	type->dimension = PyInt_AsLong(PyTuple_GetItem(value, 2));
	return 1;
    }
    else
    {
	if (PyString_Check(key))
	{
	    char pad[1024];
	    char *keystr;
	    keystr = PyString_AsString(key);
	    sprintf(pad, "Undeclared token: '%s'", keystr);
	    PyErr_SetString(terry_error, pad);
	}
	else
	{
	    PyErr_SetString(terry_error, "Undeclared token");
	}
	return 0;
    }
}

int
terry_sizeof(struct type *type)
{
    /* NOTE: This had better be in the same order as the enum in terry.h */
    int sizes[] = {
	sizeof(RtBoolean),
	sizeof(RtInt),
	sizeof(RtFloat),
	sizeof(RtString),
	sizeof(RtToken),
	sizeof(RtPoint),
	sizeof(RtColor),
	sizeof(RtMatrix),
	sizeof(RtBasis),
	sizeof(RtBound),
	sizeof(RtLightHandle),
	sizeof(RtObjectHandle),
	sizeof(RtFilterFunc),
	sizeof(RtFunc),
	sizeof(RtPointer)
    };

    if (type->base == T_COLOR)
	return sizeof(float) * terry_color_samples;
    else
	return sizes[(int)type->base];
}

static int
convert_int(PyObject *o, RtInt *data)
{
    if (PyInt_Check(o))
    {
	*data = PyInt_AsLong(o);
	return 1;
    }
    else
    {
	PyErr_SetString(PyExc_TypeError, "Expected integer");
	return 0;
    }
}

static int
convert_float(PyObject *o, RtFloat *data)
{
    if (PyFloat_Check(o))
    {
	*data = PyFloat_AsDouble(o);
	return 1;
    }
    else if (PyInt_Check(o))
    {
	*data = PyInt_AsLong(o);
	return 1;
    }
    else
    {
	PyErr_SetString(PyExc_TypeError, "Exptected float");
	return 0;
    }
}

static int
convert_handle(PyObject *o, void **data)
{
    if (PyInt_Check(o))
    {
	*data = (void*)PyInt_AsLong(o);
	return 1;
    }
    else
    {
	PyErr_SetString(PyExc_TypeError, "Expected handle");
	return 0;
    }
}

static int
convert_point(PyObject *o, RtFloat *data)
{
    int j;
    PyObject *component;

    if (PyTuple_Check(o))
    {	
	for (j = 0; j < 3; j++)
	{
	    component = PyTuple_GetItem(o, j);
	    convert_float(component, &data[j]);
	}
	return 1;
    }
    else
    {
	PyErr_SetString(PyExc_TypeError, "Exptected point");
	return 0;
    }
}

static int
convert_color(PyObject *o, RtFloat *data)
{
    int j;
    PyObject *component;

    if (PyTuple_Check(o) && PyTuple_Size(o) == terry_color_samples)
    {	
	for (j = 0; j < terry_color_samples; j++)
	{
	    component = PyTuple_GetItem(o, j);
	    convert_float(component, &data[j]);
	}
	return 1;
    }
    else
    {
	PyErr_SetString(PyExc_TypeError, "Exptected colour");
	return 0;
    }
}

static int
convert_float_func(PyObject *o,  RtFilterFunc *data)
{
    /* FIXME: Shouldn't this be convert_filter_func? */

    const char *name = PyString_AsString(o);
    if (!strcmp(name, "gaussian"))
	*data = RiGaussianFilter;
    else if (!strcmp(name, "box"))
	*data = RiBoxFilter;
    else if (!strcmp(name, "triangle"))
	*data = RiTriangleFilter;
    else if (!strcmp(name, "catmullrom"))
	*data = RiCatmullRomFilter;
    else /* FIXME: Error handling */
	*data = RiSincFilter;
    return 1;
}

static int
convert_matrix(PyObject *o, RtFloat *data)
{
    int i;
    PyObject *component;

    if (PyList_Check(o) && PyList_Size(o) == 16)
    {
	for (i = 0; i < 16; i++)
	{
	    component = PyList_GetItem(o, i);
	    convert_float(component, &data[i]);
	}

	return 1;
    }
    else 
    {
	PyErr_SetString(PyExc_TypeError, "Expected transformation matrix");
	return 0;
    }
}

static int
convert_bound(PyObject *o, RtFloat *data)
{
    int i;
    PyObject *component;

    if (PyList_Check(o) && PyList_Size(o) == 6)
    {
	for (i = 0; i < 6; i++)
	{
	    component = PyList_GetItem(o, i);
	    convert_float(component, &data[i]);
	}

	return 1;
    }
    else
    {
	PyErr_SetString(PyExc_TypeError, "Expected bound");
	return 0;
    }
}    

static int
convert_basis(PyObject *o, RtFloat *data)
{
    char *name;
    RtBasis *basis;
    basis = 0;
    if (PyString_Check(o))
    {
	name = PyString_AsString(o);
	if (!strcmp(name, "bezier"))
	    basis = &RiBezierBasis;
	else if (!strcmp(name, "bspline"))
	    basis = &RiBSplineBasis;
	else if (!strcmp(name, "catmullrom"))
	    basis = &RiCatmullRomBasis;
	else if (!strcmp(name, "hermite"))
	    basis = &RiHermiteBasis;
	else if (!strcmp(name, "power"))
	    basis = &RiPowerBasis;

	if (basis)
	{
	    memcpy(basis, data, sizeof(RtBasis));
	    return 1;
	}
	else
	{
	    PyErr_SetString(PyExc_TypeError, "Bad basis name");
	    return 0;
	}
    }
    else
	return convert_matrix(o, data);
}

static int
convert_single_value(PyObject *o, struct type *type, void *data)
{
    switch (type->base)
    {
      case T_BOOLEAN:
	if (PyInt_Check(o))
	    (*(RtBoolean*)data = PyInt_AsLong(o));
	else
	{
	    PyErr_SetString(PyExc_TypeError, "Expected boolean");
	    return 0;
	}
	break;
      case T_INTEGER:
	return convert_int(o, data);
      case T_FLOAT:
	return convert_float(o, data);
      case T_TOKEN:
	if (PyString_Check(o))
	    (*(RtToken*)data = PyString_AsString(o));
	else
	{
	    PyErr_SetString(PyExc_TypeError, "Expected token");
	    return 0;
	}
	break;
      case T_STRING:
	if (PyString_Check(o))
	    (*(RtString*)data = PyString_AsString(o));
	else
	{
	    PyErr_SetString(PyExc_TypeError, "Expected string");
	    return 0;
	}
	break;
      case T_POINT:
	return convert_point(o, data);
      case T_COLOR:
	return convert_color(o, data);
      case T_MATRIX:
	return convert_matrix(o, data);
      case T_BASIS:
	return convert_basis(o, data);
      case T_BOUND:
	return convert_bound(o, data);
      case T_POINTER:
      case T_FUNC:
	/* 
	 * These are only used in the RiProcedural call, which is currently
	 * not supported
	 */
	abort();
      case T_OBJECTHANDLE:
      case T_LIGHTHANDLE:
	return convert_handle(o, data);
      case T_FLOATFUNC:
	return convert_float_func(o, data);
    }
    return 1;
}

static int
convert_possible_array(PyObject *o, struct type *type, void *data)
{
    struct type nt;
    int j;
    int element_size;
    PyObject *component;

    element_size = terry_sizeof(type);

    if (type->dimension == 0)
    {
	return convert_single_value(o, type, data);
    }
    else
    {
	if (!PyList_Check(o))
	{
	    PyErr_SetString(PyExc_TypeError, "Expected list");
	    return 0;
	}

	if (PyList_Size(o) != type->dimension)
	{
	    PyErr_SetString(PyExc_TypeError, "Wrong dimension of array");
	    return 0;
	}

	nt = *type;
	nt.dimension = 0;
	for (j = 0; j < type->dimension; j++)
	{
	    component = PyList_GetItem(o, j);
	    if(!convert_single_value(component, &nt,
				     (void*)((char*)data + j * element_size)))
		return 0;
	}
    }

    return 1;
}

int
terry_convert_list_of_values(PyObject *o, struct type *type, void *data,
			     int element_size)
{
    int n_elements, i;
    PyObject *component;
    
    n_elements = PyList_Size(o);
    for (i = 0; i < n_elements; i++)
    {
	component = PyList_GetItem(o, i);
	if (!convert_possible_array(component, type,
				    (void*)((char*)data + i * element_size)))
	    return 0;
    }
    return 1;
}

static void *
convert_value(PyObject *o, struct type *type)
{
    void *data;
    int element_size;
    int n_elements;
    int n_sub_elements;

    element_size = terry_sizeof(type);
    n_sub_elements = (type->dimension == 0) ? 1 : type->dimension;

    if (type->varying)
    {
	if (!PyList_Check(o))
	{
	    PyErr_SetString(PyExc_TypeError,
			    "List expected for varying parameter");
	    return 0;
	}
	n_elements = PyList_Size(o);
	data = xmalloc(n_elements * element_size * n_sub_elements);
	if (!terry_convert_list_of_values(o, type, data,
					  element_size * n_sub_elements))
	{
	    xfree(data);
	    return 0;
	}
    }
    else
    {
	data = xmalloc(element_size * n_sub_elements);
	if (!convert_possible_array(o, type, data))
	{
	    xfree(data);
	    return 0;
	}
    }

    return data;
}

int
terry_allocate_parameter_list(int nparams, RtToken ** tokens,
			      RtPointer ** values)
{
    /* FIXME: Check for errors */
    *tokens = (RtToken*)xmalloc(nparams * sizeof(RtToken));
    *values = (RtPointer*)xmalloc(nparams * sizeof(RtPointer));
    /* FIXME: This assumes that NULL really is all-bits-zero... */
    memset(*tokens, 0, nparams * sizeof(RtToken));
    memset(*values, 0, nparams * sizeof(RtPointer));
    return 1;
}

int
terry_parse_parameter_list(PyObject *dict, RtToken* tokens,
			   RtPointer *values)
{
    int dictlen;
    PyObject *key, *value;
    int dictpos, i;
    struct type type;

    if (dict == 0 || dict == Py_None)
	return 0;

    if (!PyDict_Check(dict))
    {
	PyErr_SetString(terry_error, "Parameter list is not a dictionary");
	return -1;
    }

    dictlen = PyDict_Size(dict);

    dictpos = 0;
    i = 0;
    while (PyDict_Next(dict, &dictpos, &key, &value))
    {
	if (!lookup_type(key, &type)) return 0;
	tokens[i] = PyString_AsString(key);
	values[i] = convert_value(value, &type);
	if (values[i] == 0) return 0;
	i++;
    }

    return dictlen;
}

void
terry_free_parameter_list(int n, RtToken *t, RtPointer *v)
{
    int i;

    for (i = 0; i < n; i++)
    {
	if (v[i]) xfree(v[i]);
    }
}

int
terry_parse_dict_and_kwd(PyObject *dict, PyObject *kwd,
			 int *ntoks, RtToken **tokens, RtPointer **values)
{
    int n, n1;

    n = (dict ? PyDict_Size(dict) : 0) + (kwd ? PyDict_Size(kwd) : 0);
    if (ntoks) *ntoks = n;
    terry_allocate_parameter_list(n, tokens, values);
    n1 = 0;
    if (dict && (n1 = terry_parse_parameter_list(dict, *tokens, *values)) < 0)
    {
	terry_free_parameter_list(n, *tokens, *values);
	return 0;
    }
    if (kwd && terry_parse_parameter_list(kwd, *tokens + n1, *values + n1) < 0)
    {
	terry_free_parameter_list(n, *tokens, *values);
	return 0;
    }

    return 1;
}

/*
 * Parse the ARGS tuple, according to ARGSYNTAX. If VAR is not null.
 * the function has a token-value dictionary as it's last argument,
 * which will be put where var points.
 */

int
terry_parse_args(int nformal, PyObject *args, PyObject **var,
		 struct type *argsyntax, ...)
{
    int i, nargs;
    void *argptr;
    va_list argv;

    if (var) *var = 0;

    va_start(argv, argsyntax);

    nargs = PyTuple_Size(args);
    if (nargs < nformal)
    {
	PyErr_SetString(terry_error, "Too few arguments");
	va_end(argv);
	return 0;
    }

    for (i = 0; i < nformal; i++)
    {
	argptr = va_arg(argv, void*);
	if (!convert_single_value(PyTuple_GetItem(args, i), &argsyntax[i],
				  argptr))
	{
	    va_end(argv);
	    return 0;
	}
    }

    va_end(argv);

    if ((!var && nargs > nformal)
	|| (var && (nargs > nformal + 1)))
    {
	PyErr_SetString(terry_error, "Too many arguments");
	return 0;
    }
    else if (var && nargs == nformal + 1)
    {
	*var = PyTuple_GetItem(args, nargs - 1);
    }

    return 1;
}
